# Microsoft Azure DevOps

## Table of Contents 

  - [Getting Started](#getting-started)
    - [Helpful Background Information](#helpful-background-information)
    - [Prerequisites](#prerequisites)
    - [How to Install](#how-to-install)
    - [Testing](#testing)
  - [Configuration](#configuration)
    - [Example Properties](#example-properties)
    - [Connection Properties](#connection-properties)
    - [Authentication Properties](#authentication-properties)
    - [Healthcheck Properties](#healthcheck-properties)
    - [Request Properties](#request-properties)
    - [SSL Properties](#ssl-properties)
    - [Throttle Properties](#throttle-properties)
    - [Proxy Properties](#proxy-properties)
    - [Mongo Properties](#mongo-properties)
    - [Device Broker Properties](#device-broker-properties)
  - [Using this Adapter](#using-this-adapter)
    - [Generic Adapter Calls](#generic-adapter-calls)
    - [Adapter Cache Calls](#adapter-cache-calls)
    - [Adapter Broker Calls](#adapter-broker-calls)
    - [Specific Adapter Calls](#specific-adapter-calls)
    - [Authentication](#authentication)
  - [Additional Information](#additional-information)
    - [Enhancements](#enhancements)
    - [Contributing](#contributing)
    - [Helpful Links](#helpful-links)
    - [Node Scripts](#node-scripts)
  - [Troubleshoot](#troubleshoot)
    - [Connectivity Issues](#connectivity-issues)
    - [Functional Issues](#functional-issues)

## Getting Started

These instructions will help you get a copy of the project on your local machine for development and testing. Reading this section is also helpful for deployments as it provides you with pertinent information on prerequisites and properties.

### Helpful Background Information

There is <a href="https://docs.itential.com/opensource/docs/adapters" target="_blank">Adapter documentation available on the Itential Documentation Site</a>. This documentation includes information and examples that are helpful for:

```text
Authentication
IAP Service Instance Configuration
Code Files
Endpoint Configuration (Action & Schema)
Mock Data
Adapter Generic Methods
Headers
Security
Linting and Testing
Build an Adapter
Troubleshooting an Adapter
```

Others will be added over time.
Want to build a new adapter? Use the <a href="https://adapters.itential.io" target="_blank">Itential Adapter Builder</a>

### Prerequisites

The following is a list of required packages for installation on the system the adapter will run on:

```text
Node.js
npm
Git
```

The following list of packages are required for Itential opensource adapters or custom adapters that have been built utilizing the Itential Adapter Builder. You can install these packages by running npm install inside the adapter directory.

<table border="1" class="bordered-table">
  <tr>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Package</span></th>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Description</span></th>
  </tr>
  <tr>
    <td style="padding:15px">@itentialopensource/adapter-utils</td>
    <td style="padding:15px">Runtime library classes for all adapters;  includes request handling, connection, authentication throttling, and translation.</td>
  </tr>
  <tr>
    <td style="padding:15px">ajv</td>
    <td style="padding:15px">Required for validation of adapter properties to integrate with Azure_devops.</td>
  </tr>
  <tr>
    <td style="padding:15px">axios</td>
    <td style="padding:15px">Utilized by the node scripts that are included with the adapter; helps to build and extend the functionality.</td>
  </tr>
  <tr>
    <td style="padding:15px">commander</td>
    <td style="padding:15px">Utilized by the node scripts that are included with the adapter; helps to build and extend the functionality.</td>
  </tr>
  <tr>
    <td style="padding:15px">dns-lookup-promise</td>
    <td style="padding:15px">Utilized by the node scripts that are included with the adapter; helps to build and extend the functionality.</td>
  </tr>
  <tr>
    <td style="padding:15px">fs-extra</td>
    <td style="padding:15px">Utilized by the node scripts that are included with the adapter; helps to build and extend the functionality.</td>
  </tr>
  <tr>
    <td style="padding:15px">mocha</td>
    <td style="padding:15px">Testing library that is utilized by some of the node scripts that are included with the adapter.</td>
  </tr>
  <tr>
    <td style="padding:15px">mocha-param</td>
    <td style="padding:15px">Testing library that is utilized by some of the node scripts that are included with the adapter.</td>
  </tr>
  <tr>
    <td style="padding:15px">mongodb</td>
    <td style="padding:15px">Utilized by the node scripts that are included with the adapter; helps to build and extend the functionality.</td>
  </tr>
  <tr>
    <td style="padding:15px">nyc</td>
    <td style="padding:15px">Testing coverage library that is utilized by some of the node scripts that are included with the adapter.</td>
  </tr>
  <tr>
    <td style="padding:15px">ping</td>
    <td style="padding:15px">Utilized by the node scripts that are included with the adapter; helps to build and extend the functionality.</td>
  </tr>
  <tr>
    <td style="padding:15px">readline-sync</td>
    <td style="padding:15px">Utilized by the node script that comes with the adapter;  helps to test unit and integration functionality.</td>
  </tr>
  <tr>
    <td style="padding:15px">semver</td>
    <td style="padding:15px">Utilized by the node scripts that are included with the adapter; helps to build and extend the functionality.</td>
  </tr>
  <tr>
    <td style="padding:15px">winston</td>
    <td style="padding:15px">Utilized by the node scripts that are included with the adapter; helps to build and extend the functionality.</td>
  </tr>
</table>
<br>

If you are developing and testing a custom adapter, or have testing capabilities on an Itential opensource adapter, you will need to install these packages as well.

```text
chai
eslint
eslint-config-airbnb-base
eslint-plugin-import
eslint-plugin-json
testdouble
```

### How to Install

1. Set up the name space location in your IAP node_modules.

```bash
cd /opt/pronghorn/current/node_modules (* could be in a different place)
if the @itentialopensource directory does not exist, create it:
    mkdir @itentialopensource
```

2. Clone/unzip/tar the adapter into your IAP environment.

```bash
cd \@itentialopensource
git clone git@gitlab.com:\@itentialopensource/adapters/adapter-azure_devops
or
unzip adapter-azure_devops.zip
or
tar -xvf adapter-azure_devops.tar
```

3. Run the adapter install script.

```bash
cd adapter-azure_devops
npm install
npm run lint:errors
npm run test
```

4. Restart IAP

```bash
systemctl restart pronghorn
```

5. Create an adapter service instance configuration in IAP Admin Essentials GUI

6. Copy the properties from the sampleProperties.json and paste them into the service instance configuration in the inner/second properties field.

7. Change the adapter service instance configuration (host, port, credentials, etc) in IAP Admin Essentials GUI


For an easier install of the adapter use npm run adapter:install, it will install the adapter in IAP. Please note that it can be dependent on where the adapter is installed and on the version of IAP so it is subject to fail. If using this, you can replace step 3-5 above with these:

3. Install adapter dependencies and check the adapter.

```bash
cd adapter-azure_devops
npm run adapter:install
```

4. Restart IAP

```bash
systemctl restart pronghorn
```

5. Change the adapter service instance configuration (host, port, credentials, etc) in IAP Admin Essentials GUI


### Testing

Mocha is generally used to test all Itential Opensource Adapters. There are unit tests as well as integration tests performed. Integration tests can generally be run as standalone using mock data and running the adapter in stub mode, or as integrated. When running integrated, every effort is made to prevent environmental failures, however there is still a possibility.

#### Unit Testing

Unit Testing includes testing basic adapter functionality as well as error conditions that are triggered in the adapter prior to any integration. There are two ways to run unit tests. The prefered method is to use the testRunner script; however, both methods are provided here.

```bash
node utils/testRunner --unit

npm run test:unit
npm run test:baseunit
```

To add new unit tests, edit the `test/unit/adapterTestUnit.js` file. The tests that are already in this file should provide guidance for adding additional tests.

#### Integration Testing - Standalone

Standalone Integration Testing requires mock data to be provided with the entities. If this data is not provided, standalone integration testing will fail. When the adapter is set to run in stub mode (setting the stub property to true), the adapter will run through its code up to the point of making the request. It will then retrieve the mock data and return that as if it had received that data as the response from Azure_devops. It will then translate the data so that the adapter can return the expected response to the rest of the Itential software. Standalone is the default integration test.

Similar to unit testing, there are two ways to run integration tests. Using the testRunner script is better because it prevents you from having to edit the test script; it will also resets information after testing is complete so that credentials are not saved in the file.

```bash
node utils/testRunner
  answer no at the first prompt

npm run test:integration
```

To add new integration tests, edit the `test/integration/adapterTestIntegration.js` file. The tests that are already in this file should provide guidance for adding additional tests.

#### Integration Testing

Integration Testing requires connectivity to Azure_devops. By using the testRunner script it prevents you from having to edit the integration test. It also resets the integration test after the test is complete so that credentials are not saved in the file.

> **Note**: These tests have been written as a best effort to make them work in most environments. However, the Adapter Builder often does not have the necessary information that is required to set up valid integration tests. For example, the order of the requests can be very important and data is often required for `creates` and `updates`. Hence, integration tests may have to be enhanced before they will work (integrate) with Azure_devops. Even after tests have been set up properly, it is possible there are environmental constraints that could result in test failures. Some examples of possible environmental issues are customizations that have been made within Azure_devops which change order dependencies or required data.

```bash
node utils/testRunner
answer yes at the first prompt
answer all other questions on connectivity and credentials
```

Test should also be written to clean up after themselves. However, it is important to understand that in some cases this may not be possible. In addition, whenever exceptions occur, test execution may be stopped, which will prevent cleanup actions from running. It is recommended that tests be utilized in dev and test labs only.

> **Reminder**: Do not check in code with actual credentials to systems.

## Configuration

This section defines **all** the properties that are available for the adapter, including detailed information on what each property is for. If you are not using certain capabilities with this adapter, you do not need to define all of the properties. An example of how the properties for this adapter can be used with tests or IAP are provided in the sampleProperties.

### Example Properties

```json
  "properties": {
    "host": "dev.azure.com",
    "port": 443,
    "choosepath": "",
    "base_path": "/",
    "version": "",
    "cache_location": "none",
    "encode_pathvars": true,
    "encode_queryvars": true,
    "save_metric": false,
    "stub": true,
    "protocol": "https",
    "authentication": {
      "auth_method": "static_token",
      "username": "username",
      "password": "password",
      "token": "token",
      "token_timeout": 600000,
      "token_cache": "local",
      "invalid_token_error": 401,
      "auth_field": "header.headers.Authorization",
      "auth_field_format": "Basic {b64}Basic:{token}{/b64}",
      "auth_logging": false,
      "client_id": "",
      "client_secret": "",
      "grant_type": "",
      "sensitive": [],
      "sso": {
        "protocol": "",
        "host": "",
        "port": 0
      },
      "multiStepAuthCalls": [
        {
          "name": "",
          "requestFields": {},
          "responseFields": {},
          "successfullResponseCode": 200
        }
      ]
    },
    "healthcheck": {
      "type": "none",
      "frequency": 60000,
      "query_object": {},
      "addlHeaders": {}
    },
    "throttle": {
      "throttle_enabled": false,
      "number_pronghorns": 1,
      "sync_async": "sync",
      "max_in_queue": 1000,
      "concurrent_max": 1,
      "expire_timeout": 0,
      "avg_runtime": 200,
      "priorities": [
        {
          "value": 0,
          "percent": 100
        }
      ]
    },
    "request": {
      "number_redirects": 0,
      "number_retries": 3,
      "limit_retry_error": [
        0
      ],
      "failover_codes": [],
      "attempt_timeout": 5000,
      "global_request": {
        "payload": {},
        "uriOptions": {},
        "addlHeaders": {},
        "authData": {}
      },
      "healthcheck_on_timeout": true,
      "return_raw": false,
      "archiving": false,
      "return_request": false
    },
    "proxy": {
      "enabled": false,
      "host": "",
      "port": 1,
      "protocol": "http",
      "username": "",
      "password": ""
    },
    "ssl": {
      "ecdhCurve": "",
      "enabled": false,
      "accept_invalid_cert": false,
      "ca_file": "",
      "key_file": "",
      "cert_file": "",
      "secure_protocol": "",
      "ciphers": ""
    },
    "mongo": {
      "host": "",
      "port": 0,
      "database": "",
      "username": "",
      "password": "",
      "replSet": "",
      "db_ssl": {
        "enabled": false,
        "accept_invalid_cert": false,
        "ca_file": "",
        "key_file": "",
        "cert_file": ""
      }
    },
    "devicebroker": {
      "getDevice": [
        {
          "path": "/get/devices/{id}",
          "method": "GET",
          "query": {},
          "body": {},
          "headers": {},
          "handleFailure": "ignore",
          "requestFields": {
            "id": "name"
          },
          "responseDatakey": "",
          "responseFields": {
            "name": "host",
            "ostype": "os",
            "ostypePrefix": "system-",
            "ipaddress": "attributes.ipaddr",
            "port": "443"
          }
        }
      ],
      "getDevicesFiltered": [
        {
          "path": "/get/devices",
          "method": "GET",
          "pagination": {
            "offsetVar": "",
            "limitVar": "",
            "incrementBy": "limit",
            "requestLocation": "query"
          },
          "query": {},
          "body": {},
          "headers": {},
          "handleFailure": "ignore",
          "requestFields": {},
          "responseDatakey": "",
          "responseFields": {
            "name": "host",
            "ostype": "os",
            "ostypePrefix": "system-",
            "ipaddress": "attributes.ipaddr",
            "port": "443"
          }
        }
      ],
      "isAlive": [
        {
          "path": "/get/devices/{id}/status",
          "method": "GET",
          "query": {},
          "body": {},
          "headers": {},
          "handleFailure": "ignore",
          "requestFields": {
            "id": "name"
          },
          "responseDatakey": "",
          "responseFields": {
            "status": "status",
            "statusValue": "online"
          }
        }
      ],
      "getConfig": [
        {
          "path": "/get/devices/{id}/configPart1",
          "method": "GET",
          "query": {},
          "body": {},
          "headers": {},
          "handleFailure": "ignore",
          "requestFields": {
            "id": "name"
          },
          "responseDatakey": "",
          "responseFields": {}
        }
      ],
      "getCount": [
        {
          "path": "/get/devices",
          "method": "GET",
          "query": {},
          "body": {},
          "headers": {},
          "handleFailure": "ignore",
          "requestFields": {},
          "responseDatakey": "",
          "responseFields": {}
        }
      ]
    },
    "cache": {
      "enabled": false,
      "entities": [
        {
          "entityType": "",
          "frequency": 1440,
          "flushOnFail": false,
          "limit": 1000,
          "retryAttempts": 5,
          "sort": true,
          "populate": [
            {
              "path": "",
              "method": "GET",
              "pagination": {
                "offsetVar": "",
                "limitVar": "",
                "incrementBy": "limit",
                "requestLocation": "query"
              },
              "query": {},
              "body": {},
              "headers": {},
              "handleFailure": "ignore",
              "requestFields": {},
              "responseDatakey": "",
              "responseFields": {}
            }
          ],
          "cachedTasks": [
            {
              "name": "",
              "filterField": "",
              "filterLoc": ""
            }
          ]
        }
      ]
    }
  }
```

### Connection Properties

These base properties are used to connect to Azure_devops upon the adapter initially coming up. It is important to set these properties appropriately.

<table border="1" class="bordered-table">
  <tr>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Property</span></th>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Description</span></th>
  </tr>
  <tr>
    <td style="padding:15px">host</td>
    <td style="padding:15px">Required. A fully qualified domain name or IP address.</td>
  </tr>
  <tr>
    <td style="padding:15px">port</td>
    <td style="padding:15px">Required. Used to connect to the server.</td>
  </tr>
  <tr>
    <td style="padding:15px">base_path</td>
    <td style="padding:15px">Optional. Used to define part of a path that is consistent for all or most endpoints. It makes the URIs easier to use and maintain but can be overridden on individual calls. An example **base_path** might be `/rest/api`. Default is ``.</td>
  </tr>
  <tr>
    <td style="padding:15px">version</td>
    <td style="padding:15px">Optional. Used to set a global version for action endpoints. This makes it faster to update the adapter when endpoints change. As with the base-path, version can be overridden on individual endpoints. Default is ``.</td>
  </tr>
  <tr>
    <td style="padding:15px">cache_location</td>
    <td style="padding:15px">Optional. Used to define where the adapter cache is located. The cache is used to maintain an entity list to improve performance. Storage locally is lost when the adapter is restarted. Storage in Redis is preserved upon adapter restart. Default is none which means no caching of the entity list.</td>
  </tr>
  <tr>
    <td style="padding:15px">encode_pathvars</td>
    <td style="padding:15px">Optional. Used to tell the adapter to encode path variables or not. The default behavior is to encode them so this property can be used to stop that behavior.</td>
  </tr>
  <tr>
    <td style="padding:15px">encode_queryvars</td>
    <td style="padding:15px">Optional. Used to tell the adapter to encode query parameters or not. The default behavior is to encode them so this property can be used to stop that behavior.</td>
  </tr>
  <tr>
    <td style="padding:15px">save_metric</td>
    <td style="padding:15px">Optional. Used to tell the adapter to save metric information (this does not impact metrics returned on calls). This allows the adapter to gather metrics over time. Metric data can be stored in a database or on the file system.</td>
  </tr>
  <tr>
    <td style="padding:15px">stub</td>
    <td style="padding:15px">Optional. Indicates whether the stub should run instead of making calls to Azure_devops (very useful during basic testing). Default is false (which means connect to Azure_devops).</td>
  </tr>
  <tr>
    <td style="padding:15px">protocol</td>
    <td style="padding:15px">Optional. Notifies the adapter whether to use HTTP or HTTPS. Default is HTTP.</td>
  </tr>
</table>
<br>

A connectivity check tells IAP the adapter has loaded successfully.

### Authentication Properties

The following properties are used to define the authentication process to Azure_devops.

>**Note**: Depending on the method that is used to authenticate with Azure_devops, you may not need to set all of the authentication properties.

<table border="1" class="bordered-table">
  <tr>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Property</span></th>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Description</span></th>
  </tr>
  <tr>
    <td style="padding:15px">auth_method</td>
    <td style="padding:15px">Required. Used to define the type of authentication currently supported. Authentication methods currently supported are: `basic user_password`, `static_token`, `request_token`, and `no_authentication`.</td>
  </tr>
  <tr>
    <td style="padding:15px">username</td>
    <td style="padding:15px">Used to authenticate with Azure_devops on every request or when pulling a token that will be used in subsequent requests.</td>
  </tr>
  <tr>
    <td style="padding:15px">password</td>
    <td style="padding:15px">Used to authenticate with Azure_devops on every request or when pulling a token that will be used in subsequent requests.</td>
  </tr>
  <tr>
    <td style="padding:15px">token</td>
    <td style="padding:15px">Defines a static token that can be used on all requests. Only used with `static_token` as an authentication method (auth\_method).</td>
  </tr>
  <tr>
    <td style="padding:15px">invalid_token_error</td>
    <td style="padding:15px">Defines the HTTP error that is received when the token is invalid. Notifies the adapter to pull a new token and retry the request. Default is 401.</td>
  </tr>
  <tr>
    <td style="padding:15px">token_timeout</td>
    <td style="padding:15px">Defines how long a token is valid. Measured in milliseconds. Once a dynamic token is no longer valid, the adapter has to pull a new token. If the token_timeout is set to -1, the adapter will pull a token on every request to Azure_devops. If the timeout_token is 0, the adapter will use the expiration from the token response to determine when the token is no longer valid.</td>
  </tr>
  <tr>
    <td style="padding:15px">token_cache</td>
    <td style="padding:15px">Used to determine where the token should be stored (local memory or in Redis).</td>
  </tr>
  <tr>
    <td style="padding:15px">auth_field</td>
    <td style="padding:15px">Defines the request field the authentication (e.g., token are basic auth credentials) needs to be placed in order for the calls to work.</td>
  </tr>
  <tr>
    <td style="padding:15px">auth_field_format</td>
    <td style="padding:15px">Defines the format of the auth\_field. See examples below. Items enclosed in {} inform the adapter to perofrm an action prior to sending the data. It may be to replace the item with a value or it may be to encode the item.</td>
  </tr>
  <tr>
    <td style="padding:15px">auth_logging</td>
    <td style="padding:15px">Setting this true will add some additional logs but this should only be done when trying to debug an issue as certain credential information may be logged out when this is true.</td>
  </tr>
  <tr>
    <td style="padding:15px">client_id</td>
    <td style="padding:15px">Provide a client id when needed, this is common on some types of OAuth.</td>
  </tr>
  <tr>
    <td style="padding:15px">client_secret</td>
    <td style="padding:15px">Provide a client secret when needed, this is common on some types of OAuth.</td>
  </tr>
  <tr>
    <td style="padding:15px">grant_type</td>
    <td style="padding:15px">Provide a grant type when needed, this is common on some types of OAuth.</td>
  </tr>
</table>
<br>

#### Examples of authentication field format

```json
"{token}"
"Token {token}"
"{username}:{password}"
"Basic {b64}{username}:{password}{/b64}"
```

### Healthcheck Properties

The healthcheck properties defines the API that runs the healthcheck to tell the adapter that it can reach Azure_devops. There are currently three types of healthchecks.

- None - Not recommended. Adapter will not run a healthcheck. Consequently, unable to determine before making a request if the adapter can reach Azure_devops.
- Startup - Adapter will check for connectivity when the adapter initially comes up, but it will not check afterwards.
- Intermittent - Adapter will check connectivity to Azure_devops at a frequency defined in the `frequency` property.

<table border="1" class="bordered-table">
  <tr>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Property</span></th>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Description</span></th>
  </tr>
  <tr>
    <td style="padding:15px">type</td>
    <td style="padding:15px">Required. The type of health check to run.</td>
  </tr>
  <tr>
    <td style="padding:15px">frequency</td>
    <td style="padding:15px">Required if intermittent. Defines how often the health check should run. Measured in milliseconds. Default is 300000.</td>
  </tr>
  <tr>
    <td style="padding:15px">query_object</td>
    <td style="padding:15px">Query parameters to be added to the adapter healthcheck call.</td>
  </tr>
</table>
<br>

### Request Properties

The request section defines properties to help handle requests.

<table border="1" class="bordered-table">
  <tr>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Property</span></th>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Description</span></th>
  </tr>
  <tr>
    <td style="padding:15px">number_redirects</td>
    <td style="padding:15px">Optional. Tells the adapter that the request may be redirected and gives it a maximum number of redirects to allow before returning an error. Default is 0 - no redirects.</td>
  </tr>
  <tr>
    <td style="padding:15px">number_retries</td>
    <td style="padding:15px">Tells the adapter how many times to retry a request that has either aborted or reached a limit error before giving up and returning an error.</td>
  </tr>
  <tr>
    <td style="padding:15px">limit_retry_error</td>
    <td style="padding:15px">Optional. Can be either an integer or an array. Indicates the http error status number to define that no capacity was available and, after waiting a short interval, the adapter can retry the request. If an array is provvided, the array can contain integers or strings. Strings in the array are used to define ranges (e.g. "502-506"). Default is [0].</td>
  </tr>
  <tr>
    <td style="padding:15px">failover_codes</td>
    <td style="padding:15px">An array of error codes for which the adapter will send back a failover flag to IAP so that the Platform can attempt the action in another adapter.</td>
  </tr>
  <tr>
    <td style="padding:15px">attempt_timeout</td>
    <td style="padding:15px">Optional. Tells how long the adapter should wait before aborting the attempt. On abort, the adapter will do one of two things: 1) return the error; or 2) if **healthcheck\_on\_timeout** is set to true, it will abort the request and run a Healthcheck until it re-establishes connectivity to Azure_devops, and then will re-attempt the request that aborted. Default is 5000 milliseconds.</td>
  </tr>
  <tr>
    <td style="padding:15px">global_request</td>
    <td style="padding:15px">Optional. This is information that the adapter can include in all requests to the other system. This is easier to define and maintain than adding this information in either the code (adapter.js) or the action files.</td>
  </tr>
  <tr>
    <td style="padding:15px">global_request -> payload</td>
    <td style="padding:15px">Optional. Defines any information that should be included on all requests sent to the other system that have a payload/body.</td>
  </tr>
  <tr>
    <td style="padding:15px">global_request -> uriOptions</td>
    <td style="padding:15px">Optional. Defines any information that should be sent as untranslated  query options (e.g. page, size) on all requests to the other system.</td>
  </tr>
  <tr>
    <td style="padding:15px">global_request -> addlHeaders</td>
    <td style="padding:15px">Optioonal. Defines any headers that should be sent on all requests to the other system.</td>
  </tr>
  <tr>
    <td style="padding:15px">global_request -> authData</td>
    <td style="padding:15px">Optional. Defines any additional authentication data used to authentice with the other system. This authData needs to be consistent on every request.</td>
  </tr>
  <tr>
    <td style="padding:15px">healthcheck_on_timeout</td>
    <td style="padding:15px">Required. Defines if the adapter should run a health check on timeout. If set to true, the adapter will abort the request and run a health check until it re-establishes connectivity and then it will re-attempt the request.</td>
  </tr>
  <tr>
    <td style="padding:15px">return_raw</td>
    <td style="padding:15px">Optional. Tells the adapter whether the raw response should be returned as well as the IAP response. This is helpful when running integration tests to save mock data. It does add overhead to the response object so it is not ideal from production.</td>
  </tr>
  <tr>
    <td style="padding:15px">archiving</td>
    <td style="padding:15px">Optional flag. Default is false. It archives the request, the results and the various times (wait time, Azure_devops time and overall time) in the `adapterid_results` collection in MongoDB. Although archiving might be desirable, be sure to develop a strategy before enabling this capability. Consider how much to archive and what strategy to use for cleaning up the collection in the database so that it does not become too large, especially if the responses are large.</td>
  </tr>
  <tr>
    <td style="padding:15px">return_request</td>
    <td style="padding:15px">Optional flag. Default is false. Will return the actual request that is made including headers. This should only be used during debugging issues as there could be credentials in the actual request.</td>
  </tr>
</table>
<br>

### SSL Properties

The SSL section defines the properties utilized for ssl authentication with Azure_devops. SSL can work two different ways: set the `accept\_invalid\_certs` flag to true (only recommended for lab environments), or provide a `ca\_file`.

<table border="1" class="bordered-table">
  <tr>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Property</span></th>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Description</span></th>
  </tr>
  <tr>
    <td style="padding:15px">enabled</td>
    <td style="padding:15px">If SSL is required, set to true.</td>
  </tr>
  <tr>
    <td style="padding:15px">accept_invalid_certs</td>
    <td style="padding:15px">Defines if the adapter should accept invalid certificates (only recommended for lab environments). Required if SSL is enabled. Default is false.</td>
  </tr>
  <tr>
    <td style="padding:15px">ca_file</td>
    <td style="padding:15px">Defines the path name to the CA file used for SSL. If SSL is enabled and the accept invalid certifications is false, then ca_file is required.</td>
  </tr>
  <tr>
    <td style="padding:15px">key_file</td>
    <td style="padding:15px">Defines the path name to the Key file used for SSL. The key_file may be needed for some systems but it is not required for SSL.</td>
  </tr>
  <tr>
    <td style="padding:15px">cert_file</td>
    <td style="padding:15px">Defines the path name to the Certificate file used for SSL. The cert_file may be needed for some systems but it is not required for SSL.</td>
  </tr>
  <tr>
    <td style="padding:15px">secure_protocol</td>
    <td style="padding:15px">Defines the protocol (e.g., SSLv3_method) to use on the SSL request.</td>
  </tr>
  <tr>
    <td style="padding:15px">ciphers</td>
    <td style="padding:15px">Required if SSL enabled. Specifies a list of SSL ciphers to use.</td>
  </tr>
  <tr>
    <td style="padding:15px">ecdhCurve</td>
    <td style="padding:15px">During testing on some Node 8 environments, you need to set `ecdhCurve` to auto. If you do not, you will receive PROTO errors when attempting the calls. This is the only usage of this property and to our knowledge it only impacts Node 8 and 9.</td>
  </tr>
</table>
<br>

### Throttle Properties

The throttle section is used when requests to Azure_devops must be queued (throttled). All of the properties in this section are optional.

<table border="1" class="bordered-table">
  <tr>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Property</span></th>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Description</span></th>
  </tr>
  <tr>
    <td style="padding:15px">throttle_enabled</td>
    <td style="padding:15px">Default is false. Defines if the adapter should use throttling or not.</td>
  </tr>
  <tr>
    <td style="padding:15px">number_pronghorns</td>
    <td style="padding:15px">Default is 1. Defines if throttling is done in a single Itential instance or whether requests are being throttled across multiple Itential instances (minimum = 1, maximum = 20). Throttling in a single Itential instance uses an in-memory queue so there is less overhead. Throttling across multiple Itential instances requires placing the request and queue information into a shared resource (e.g. database) so that each instance can determine what is running and what is next to run. Throttling across multiple instances requires additional I/O overhead.</td>
  </tr>
  <tr>
    <td style="padding:15px">sync-async</td>
    <td style="padding:15px">This property is not used at the current time (it is for future expansion of the throttling engine).</td>
  </tr>
  <tr>
    <td style="padding:15px">max_in_queue</td>
    <td style="padding:15px">Represents the maximum number of requests the adapter should allow into the queue before rejecting requests (minimum = 1, maximum = 5000). This is not a limit on what the adapter can handle but more about timely responses to requests. The default is currently 1000.</td>
  </tr>
  <tr>
    <td style="padding:15px">concurrent_max</td>
    <td style="padding:15px">Defines the number of requests the adapter can send to Azure_devops at one time (minimum = 1, maximum = 1000). The default is 1 meaning each request must be sent to Azure_devops in a serial manner.</td>
  </tr>
  <tr>
    <td style="padding:15px">expire_timeout</td>
    <td style="padding:15px">Default is 0. Defines a graceful timeout of the request session. After a request has completed, the adapter will wait additional time prior to sending the next request. Measured in milliseconds (minimum = 0, maximum = 60000).</td>
  </tr>
  <tr>
    <td style="padding:15px">average_runtime</td>
    <td style="padding:15px">Represents the approximate average of how long it takes Azure_devops to handle each request. Measured in milliseconds (minimum = 50, maximum = 60000). Default is 200. This metric has performance implications. If the runtime number is set too low, it puts extra burden on the CPU and memory as the requests will continually try to run. If the runtime number is set too high, requests may wait longer than they need to before running. The number does not need to be exact but your throttling strategy depends heavily on this number being within reason. If averages range from 50 to 250 milliseconds you might pick an average run-time somewhere in the middle so that when Azure_devops performance is exceptional you might run a little slower than you might like, but when it is poor you still run efficiently.</td>
  </tr>
  <tr>
    <td style="padding:15px">priorities</td>
    <td style="padding:15px">An array of priorities and how to handle them in relation to the throttle queue. Array of objects that include priority value and percent of queue to put the item ex { value: 1, percent: 10 }</td>
  </tr>
</table>
<br>

### Proxy Properties

The proxy section defines the properties to utilize when Azure_devops is behind a proxy server.

<table border="1" class="bordered-table">
  <tr>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Property</span></th>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Description</span></th>
  </tr>
  <tr>
    <td style="padding:15px">enabled</td>
    <td style="padding:15px">Required. Default is false. If Azure_devops is behind a proxy server, set enabled flag to true.</td>
  </tr>
  <tr>
    <td style="padding:15px">host</td>
    <td style="padding:15px">Host information for the proxy server. Required if `enabled` is true.</td>
  </tr>
  <tr>
    <td style="padding:15px">port</td>
    <td style="padding:15px">Port information for the proxy server. Required if `enabled` is true.</td>
  </tr>
  <tr>
    <td style="padding:15px">protocol</td>
    <td style="padding:15px">The protocol (i.e., http, https, etc.) used to connect to the proxy. Default is http.</td>
  </tr>
  <tr>
    <td style="padding:15px">username</td>
    <td style="padding:15px">If there is authentication for the proxy, provide the username here.</td>
  </tr>
  <tr>
    <td style="padding:15px">password</td>
    <td style="padding:15px">If there is authentication for the proxy, provide the password here.</td>
  </tr>
</table>
<br>

### Mongo Properties

The mongo section defines the properties used to connect to a Mongo database. Mongo can be used for throttling as well as to persist metric data. If not provided, metrics will be stored in the file system.

<table border="1" class="bordered-table">
  <tr>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Property</span></th>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Description</span></th>
  </tr>
  <tr>
    <td style="padding:15px">host</td>
    <td style="padding:15px">Optional. Host information for the mongo server.</td>
  </tr>
  <tr>
    <td style="padding:15px">port</td>
    <td style="padding:15px">Optional. Port information for the mongo server.</td>
  </tr>
  <tr>
    <td style="padding:15px">database</td>
    <td style="padding:15px">Optional. The database for the adapter to use for its data.</td>
  </tr>
  <tr>
    <td style="padding:15px">username</td>
    <td style="padding:15px">Optional. If credentials are required to access mongo, this is the user to login as.</td>
  </tr>
  <tr>
    <td style="padding:15px">password</td>
    <td style="padding:15px">Optional. If credentials are required to access mongo, this is the password to login with.</td>
  </tr>
  <tr>
    <td style="padding:15px">replSet</td>
    <td style="padding:15px">Optional. If the database is set up to use replica sets, define it here so it can be added to the database connection.</td>
  </tr>
  <tr>
    <td style="padding:15px">db_ssl</td>
    <td style="padding:15px">Optional. Contains information for SSL connectivity to the database.</td>
  </tr>
  <tr>
    <td style="padding:15px">db_ssl -> enabled</td>
    <td style="padding:15px">If SSL is required, set to true.</td>
  </tr>
  <tr>
    <td style="padding:15px">db_ssl -> accept_invalid_cert</td>
    <td style="padding:15px">Defines if the adapter should accept invalid certificates (only recommended for lab environments). Required if SSL is enabled. Default is false.</td>
  </tr>
  <tr>
    <td style="padding:15px">db_ssl -> ca_file</td>
    <td style="padding:15px">Defines the path name to the CA file used for SSL. If SSL is enabled and the accept invalid certifications is false, then ca_file is required.</td>
  </tr>
  <tr>
    <td style="padding:15px">db_ssl -> key_file</td>
    <td style="padding:15px">Defines the path name to the Key file used for SSL. The key_file may be needed for some systems but it is not required for SSL.</td>
  </tr>
  <tr>
    <td style="padding:15px">db_ssl -> cert_file</td>
    <td style="padding:15px">Defines the path name to the Certificate file used for SSL. The cert_file may be needed for some systems but it is not required for SSL.</td>
  </tr>
</table>
<br>

### Device Broker Properties

The device broker section defines the properties used integrate Azure_devops to the device broker. Each broker call is represented and has an array of calls that can be used to build the response. This describes the calls and then the fields which are available in the calls.

<table border="1" class="bordered-table">
  <tr>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Property</span></th>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Description</span></th>
  </tr>
  <tr>
    <td style="padding:15px">getDevice</td>
    <td style="padding:15px">The array of calls used to get device details for the broker</td>
  </tr>
  <tr>
    <td style="padding:15px">getDevicesFiltered</td>
    <td style="padding:15px">The array of calls used to get devices for the broker</td>
  </tr>
  <tr>
    <td style="padding:15px">isAlive</td>
    <td style="padding:15px">The array of calls used to get device status for the broker</td>
  </tr>
  <tr>
    <td style="padding:15px">getConfig</td>
    <td style="padding:15px">The array of calls used to get device configuration for the broker</td>
  </tr>
  <tr>
    <td style="padding:15px">getCount</td>
    <td style="padding:15px">The array of calls used to get device configuration for the broker</td>
  </tr>
  <tr>
    <td style="padding:15px">getDevice/getDevicesFiltered/isAlive/getConfig/getCount -> path</td>
    <td style="padding:15px">The path, not including the base_path and version, for making this call</td>
  </tr>
  <tr>
    <td style="padding:15px">getDevice/getDevicesFiltered/isAlive/getConfig/getCount -> method</td>
    <td style="padding:15px">The rest method for making this call</td>
  </tr>
  <tr>
    <td style="padding:15px">getDevice/getDevicesFiltered/isAlive/getConfig/getCount -> query</td>
    <td style="padding:15px">Query object containing and query parameters and their values for this call</td>
  </tr>
  <tr>
    <td style="padding:15px">getDevice/getDevicesFiltered/isAlive/getConfig/getCount -> body</td>
    <td style="padding:15px">Body object containing the payload for this call</td>
  </tr>
  <tr>
    <td style="padding:15px">getDevice/getDevicesFiltered/isAlive/getConfig/getCount -> headers</td>
    <td style="padding:15px">Header object containing the headers for this call.</td>
  </tr>
  <tr>
    <td style="padding:15px">getDevice/getDevicesFiltered/isAlive/getConfig/getCount -> handleFailure</td>
    <td style="padding:15px">Tells the adapter whether to "fail" or "ignore" failures if they occur.</td>
  </tr>
  <tr>
    <td style="padding:15px">isAlive -> statusValue</td>
    <td style="padding:15px">Tells the adapter what value to look for in the status field to determine if the device is alive.</td>
  </tr>
  <tr>
    <td style="padding:15px">getDevice/getDevicesFiltered/isAlive/getConfig -> requestFields</td>
    <td style="padding:15px">Object containing fields the adapter should send on the request and where it should get the data. The where can be from a response to a getDevicesFiltered or a static value.</td>
  </tr>
  <tr>
    <td style="padding:15px">getDevice/getDevicesFiltered/isAlive/getConfig -> responseFields</td>
    <td style="padding:15px">Object containing fields the adapter should set to send back to iap and where the value should come from in the response or request data.</td>
  </tr>
</table>
<br>


## Using this Adapter

The `adapter.js` file contains the calls the adapter makes available to the rest of the Itential Platform. The API detailed for these calls should be available through JSDOC. The following is a brief summary of the calls.

### Generic Adapter Calls

These are adapter methods that IAP or you might use. There are some other methods not shown here that might be used for internal adapter functionality.

<table border="1" class="bordered-table">
  <tr>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Method Signature</span></th>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Description</span></th>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Workflow?</span></th>
  </tr>
  <tr>
    <td style="padding:15px">connect()</td>
    <td style="padding:15px">This call is run when the Adapter is first loaded by he Itential Platform. It validates the properties have been provided correctly.</td>
    <td style="padding:15px">No</td>
  </tr>
  <tr>
    <td style="padding:15px">healthCheck(callback)</td>
    <td style="padding:15px">This call ensures that the adapter can communicate with Adapter for Microsoft Azure DevOps. The actual call that is used is defined in the adapter properties and .system entities action.json file.</td>
    <td style="padding:15px">No</td>
  </tr>
  <tr>
    <td style="padding:15px">refreshProperties(properties)</td>
    <td style="padding:15px">This call provides the adapter the ability to accept property changes without having to restart the adapter.</td>
    <td style="padding:15px">No</td>
  </tr>
  <tr>
    <td style="padding:15px">encryptProperty(property, technique, callback)</td>
    <td style="padding:15px">This call will take the provided property and technique, and return the property encrypted with the technique. This allows the property to be used in the adapterProps section for the credential password so that the password does not have to be in clear text. The adapter will decrypt the property as needed for communications with Adapter for Microsoft Azure DevOps.</td>
    <td style="padding:15px">No</td>
  </tr>
  <tr>
    <td style="padding:15px">iapUpdateAdapterConfiguration(configFile, changes, entity, type, action, callback)</td>
    <td style="padding:15px">This call provides the ability to update the adapter configuration from IAP - includes actions, schema, mockdata and other configurations.</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">iapSuspendAdapter(mode, callback)</td>
    <td style="padding:15px">This call provides the ability to suspend the adapter and either have requests rejected or put into a queue to be processed after the adapter is resumed.</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">iapUnsuspendAdapter(callback)</td>
    <td style="padding:15px">This call provides the ability to resume a suspended adapter. Any requests in queue will be processed before new requests.</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">iapGetAdapterQueue(callback)</td>
    <td style="padding:15px">This call will return the requests that are waiting in the queue if throttling is enabled.</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">iapFindAdapterPath(apiPath, callback)</td>
    <td style="padding:15px">This call provides the ability to see if a particular API path is supported by the adapter.</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">iapTroubleshootAdapter(props, persistFlag, adapter, callback)</td>
    <td style="padding:15px">This call can be used to check on the performance of the adapter - it checks connectivity, healthcheck and basic get calls.</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">iapRunAdapterHealthcheck(adapter, callback)</td>
    <td style="padding:15px">This call will return the results of a healthcheck.</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">iapRunAdapterConnectivity(callback)</td>
    <td style="padding:15px">This call will return the results of a connectivity check.</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">iapRunAdapterBasicGet(callback)</td>
    <td style="padding:15px">This call will return the results of running basic get API calls.</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">iapMoveAdapterEntitiesToDB(callback)</td>
    <td style="padding:15px">This call will push the adapter configuration from the entities directory into the Adapter or IAP Database.</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">iapDeactivateTasks(tasks, callback)</td>
    <td style="padding:15px">This call provides the ability to remove tasks from the adapter.</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">iapActivateTasks(tasks, callback)</td>
    <td style="padding:15px">This call provides the ability to add deactivated tasks back into the adapter.</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">iapExpandedGenericAdapterRequest(metadata, uriPath, restMethod, pathVars, queryData, requestBody, addlHeaders, callback)</td>
    <td style="padding:15px">This is an expanded Generic Call. The metadata object allows us to provide many new capabilities within the generic request.</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">genericAdapterRequest(uriPath, restMethod, queryData, requestBody, addlHeaders, callback)</td>
    <td style="padding:15px">This call allows you to provide the path to have the adapter call. It is an easy way to incorporate paths that have not been built into the adapter yet.</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">genericAdapterRequestNoBasePath(uriPath, restMethod, queryData, requestBody, addlHeaders, callback)</td>
    <td style="padding:15px">This call is the same as the genericAdapterRequest only it does not add a base_path or version to the call.</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">iapRunAdapterLint(callback)</td>
    <td style="padding:15px">Runs lint on the addapter and provides the information back.</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">iapRunAdapterTests(callback)</td>
    <td style="padding:15px">Runs baseunit and unit tests on the adapter and provides the information back.</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">iapGetAdapterInventory(callback)</td>
    <td style="padding:15px">This call provides some inventory related information about the adapter.</td>
    <td style="padding:15px">Yes</td>
  </tr>
</table>
<br>
  
### Adapter Cache Calls

These are adapter methods that are used for adapter caching. If configured, the adapter will cache based on the interval provided. However, you can force a population of the cache manually as well.

<table border="1" class="bordered-table">
  <tr>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Method Signature</span></th>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Description</span></th>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Workflow?</span></th>
  </tr>
  <tr>
    <td style="padding:15px">iapPopulateEntityCache(entityTypes, callback)</td>
    <td style="padding:15px">This call populates the adapter cache.</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">iapRetrieveEntitiesCache(entityType, options, callback)</td>
    <td style="padding:15px">This call retrieves the specific items from the adapter cache.</td>
    <td style="padding:15px">Yes</td>
  </tr>
</table>
<br>
  
### Adapter Broker Calls

These are adapter methods that are used to integrate to IAP Brokers. This adapter currently supports the following broker calls.

<table border="1" class="bordered-table">
  <tr>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Method Signature</span></th>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Description</span></th>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Workflow?</span></th>
  </tr>
  <tr>
    <td style="padding:15px">hasEntities(entityType, entityList, callback)</td>
    <td style="padding:15px">This call is utilized by the IAP Device Broker to determine if the adapter has a specific entity and item of the entity.</td>
    <td style="padding:15px">No</td>
  </tr>
  <tr>
    <td style="padding:15px">getDevice(deviceName, callback)</td>
    <td style="padding:15px">This call returns the details of the requested device.</td>
    <td style="padding:15px">No</td>
  </tr>
  <tr>
    <td style="padding:15px">getDevicesFiltered(options, callback)</td>
    <td style="padding:15px">This call returns the list of devices that match the criteria provided in the options filter.</td>
    <td style="padding:15px">No</td>
  </tr>
  <tr>
    <td style="padding:15px">isAlive(deviceName, callback)</td>
    <td style="padding:15px">This call returns whether the device status is active</td>
    <td style="padding:15px">No</td>
  </tr>
  <tr>
    <td style="padding:15px">getConfig(deviceName, format, callback)</td>
    <td style="padding:15px">This call returns the configuration for the selected device.</td>
    <td style="padding:15px">No</td>
  </tr>
  <tr>
    <td style="padding:15px">iapGetDeviceCount(callback)</td>
    <td style="padding:15px">This call returns the count of devices.</td>
    <td style="padding:15px">No</td>
  </tr>
</table>
<br>

### Specific Adapter Calls

Specific adapter calls are built based on the API of the Microsoft Azure DevOps. The Adapter Builder creates the proper method comments for generating JS-DOC for the adapter. This is the best way to get information on the calls.

<table border="1" class="bordered-table">
  <tr>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Method Signature</span></th>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Description</span></th>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Path</span></th>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Workflow?</span></th>
  </tr>
  <tr>
    <td style="padding:15px">pipelinesCreate(organization, body, project, apiVersion, callback)</td>
    <td style="padding:15px">Create a pipeline.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/pipelines?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pipelinesList(organization, project, orderBy, top, continuationToken, apiVersion, callback)</td>
    <td style="padding:15px">Get a list of pipelines.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/pipelines?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pipelinesGet(organization, project, pipelineId, pipelineVersion, apiVersion, callback)</td>
    <td style="padding:15px">Gets a pipeline, optionally at the specified version</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/pipelines/{pathv3}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">previewPreview(organization, body, project, pipelineId, pipelineVersion, apiVersion, callback)</td>
    <td style="padding:15px">Queues a dry run of the pipeline and returns an object containing the final yaml.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/pipelines/{pathv3}/preview?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">runsList(organization, project, pipelineId, apiVersion, callback)</td>
    <td style="padding:15px">Gets top 10000 runs for a particular pipeline.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/pipelines/{pathv3}/runs?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">runsRunPipeline(organization, body, project, pipelineId, pipelineVersion, apiVersion, callback)</td>
    <td style="padding:15px">Runs a pipeline.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/pipelines/{pathv3}/runs?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">runsGet(organization, project, pipelineId, runId, apiVersion, callback)</td>
    <td style="padding:15px">Gets a run for a particular pipeline.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/pipelines/{pathv3}/runs/{pathv4}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">artifactsGet(organization, project, pipelineId, runId, artifactName, expand = 'none', apiVersion, callback)</td>
    <td style="padding:15px">Get a specific artifact from a pipeline run</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/pipelines/{pathv3}/runs/{pathv4}/artifacts?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">logsList(organization, project, pipelineId, runId, expand = 'none', apiVersion, callback)</td>
    <td style="padding:15px">Get a list of logs from a pipeline run.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/pipelines/{pathv3}/runs/{pathv4}/logs?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">logsGet(organization, project, pipelineId, runId, logId, expand = 'none', apiVersion, callback)</td>
    <td style="padding:15px">Get a specific log from a pipeline run</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/pipelines/{pathv3}/runs/{pathv4}/logs/{pathv5}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">repositoriesGetDeletedRepositories(organization, project, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve deleted git repositories.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/deletedrepositories?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">repositoriesGetRecycleBinRepositories(organization, project, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve soft-deleted git repositories from the recycle bin.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/recycleBin/repositories?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">repositoriesDeleteRepositoryFromRecycleBin(organization, project, repositoryId, apiVersion, callback)</td>
    <td style="padding:15px">Destroy (hard delete) a soft-deleted Git repository.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/recycleBin/repositories/{pathv3}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">repositoriesRestoreRepositoryFromRecycleBin(organization, body, project, repositoryId, apiVersion, callback)</td>
    <td style="padding:15px">Recover a soft-deleted Git repository. Recently deleted repositories go into a soft-delete state fo</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/recycleBin/repositories/{pathv3}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">repositoriesCreate(organization, body, project, sourceRef, apiVersion, callback)</td>
    <td style="padding:15px">Create a git repository in a team project.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">repositoriesList(organization, project, includeLinks, includeAllUrls, includeHidden, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve git repositories.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">repositoriesDelete(organization, repositoryId, project, apiVersion, callback)</td>
    <td style="padding:15px">Delete a git repository</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">repositoriesGetRepository(organization, repositoryId, project, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve a git repository.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">repositoriesUpdate(organization, body, repositoryId, project, apiVersion, callback)</td>
    <td style="padding:15px">Updates the Git repository with either a new repo name or a new default branch.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">refsFavoritesCreate(organization, body, project, apiVersion, callback)</td>
    <td style="padding:15px">Creates a ref favorite</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/favorites/refs?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">refsFavoritesList(organization, project, repositoryId, identityId, apiVersion, callback)</td>
    <td style="padding:15px">Gets the refs favorites for a repo and an identity.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/favorites/refs?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">refsFavoritesDelete(organization, project, favoriteId, apiVersion, callback)</td>
    <td style="padding:15px">Deletes the refs favorite specified</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/favorites/refs/{pathv3}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">refsFavoritesGet(organization, project, favoriteId, apiVersion, callback)</td>
    <td style="padding:15px">Gets the refs favorite for a favorite Id.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/favorites/refs/{pathv3}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">policyConfigurationsGet(organization, project, repositoryId, refName, policyType, top, continuationToken, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve a list of policy configurations by a given set of scope/filtering criteria.

Below is a sh</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/policy/configurations?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestsGetPullRequestsByProject(organization, project, searchCriteriaCreatorId, searchCriteriaIncludeLinks, searchCriteriaRepositoryId, searchCriteriaReviewerId, searchCriteriaSourceRefName, searchCriteriaSourceRepositoryId, searchCriteriaStatus = 'notSet', searchCriteriaTargetRefName, maxCommentLength, skip, top, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve all pull requests matching a specified criteria.

Please note that description field will</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/pullrequests?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestsGetPullRequestById(organization, pullRequestId, project, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve a pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/pullrequests/{pathv3}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestsCreate(organization, body, repositoryId, project, supportsIterations, apiVersion, callback)</td>
    <td style="padding:15px">Create a pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullrequests?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestsGetPullRequests(organization, repositoryId, project, searchCriteriaCreatorId, searchCriteriaIncludeLinks, searchCriteriaRepositoryId, searchCriteriaReviewerId, searchCriteriaSourceRefName, searchCriteriaSourceRepositoryId, searchCriteriaStatus = 'notSet', searchCriteriaTargetRefName, maxCommentLength, skip, top, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve all pull requests matching a specified criteria.

Please note that description field will</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullrequests?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestsGetPullRequest(organization, repositoryId, pullRequestId, project, maxCommentLength, skip, top, includeCommits, includeWorkItemRefs, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve a pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullrequests/{pathv4}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestsUpdate(organization, body, repositoryId, pullRequestId, project, apiVersion, callback)</td>
    <td style="padding:15px">Update a pull request

These are the properties that can be updated with the API:
 - Status
 - Titl</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullrequests/{pathv4}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">annotatedTagsCreate(organization, body, project, repositoryId, apiVersion, callback)</td>
    <td style="padding:15px">Create an annotated tag.

Repositories have both a name and an identifier. Identifiers are globally</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/annotatedtags?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">annotatedTagsGet(organization, project, repositoryId, objectId, apiVersion, callback)</td>
    <td style="padding:15px">Get an annotated tag.

Repositories have both a name and an identifier. Identifiers are globally un</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/annotatedtags/{pathv4}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">blobsGetBlobsZip(organization, body, repositoryId, project, filename, apiVersion, callback)</td>
    <td style="padding:15px">Gets one or more blobs in a zip file download.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/blobs?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">blobsGetBlob(organization, repositoryId, sha1, project, download, fileName, format, resolveLfs, apiVersion, callback)</td>
    <td style="padding:15px">Get a single blob.

Repositories have both a name and an identifier. Identifiers are globally uniqu</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/blobs/{pathv4}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">cherryPicksCreate(organization, body, project, repositoryId, apiVersion, callback)</td>
    <td style="padding:15px">Cherry pick a specific commit or commits that are associated to a pull request into a new branch.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/cherryPicks?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">cherryPicksGetCherryPickForRefName(organization, project, repositoryId, refName, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve information about a cherry pick operation for a specific branch. This operation is expensi</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/cherryPicks?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">cherryPicksGetCherryPick(organization, project, cherryPickId, repositoryId, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve information about a cherry pick operation by cherry pick Id.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/cherryPicks/{pathv4}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">commitsGetPushCommits(organization, repositoryId, pushId, project, top, skip, includeLinks, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve a list of commits associated with a particular push.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/commits?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">commitsGet(organization, commitId, repositoryId, project, changeCount, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve a particular commit.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/commits/{pathv4}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">commitsGetChanges(organization, commitId, repositoryId, project, top, skip, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve changes for a particular commit.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/commits/{pathv4}/changes?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">commitsGetCommits(organization, repositoryId, project, skip, top, searchCriteriaAuthor, searchCriteriaCompareVersionVersion, searchCriteriaCompareVersionVersionOptions, searchCriteriaCompareVersionVersionType, searchCriteriaExcludeDeletes, searchCriteriaFromCommitId, searchCriteriaFromDate, searchCriteriaHistoryMode, searchCriteriaIds, searchCriteriaIncludeLinks, searchCriteriaIncludePushData, searchCriteriaIncludeUserImageUrl, searchCriteriaIncludeWorkItems, searchCriteriaItemPath, searchCriteriaItemVersionVersion, searchCriteriaItemVersionVersionOptions, searchCriteriaItemVersionVersionType, searchCriteriaShowOldestCommitsFirst, searchCriteriaToCommitId, searchCriteriaToDate, searchCriteriaUser, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve git commits for a project.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/commits?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">commitsGetCommitsBatch(organization, body, repositoryId, project, skip, top, includeStatuses, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve git commits for a project matching the search criteria</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/commitsbatch?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">statusesCreate(organization, body, commitId, repositoryId, project, apiVersion, callback)</td>
    <td style="padding:15px">Create Git commit status.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/commits/{pathv4}/statuses?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">statusesList(organization, commitId, repositoryId, project, top, skip, latestOnly, apiVersion, callback)</td>
    <td style="padding:15px">Get statuses associated with the Git commit.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/commits/{pathv4}/statuses?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">diffsGet(organization, repositoryId, project, diffCommonCommit, top, skip, baseVersion, baseVersionOptions = 'none', baseVersionType = 'branch', targetVersion, targetVersionOptions = 'none', targetVersionType = 'branch', apiVersion, callback)</td>
    <td style="padding:15px">Find the closest common commit (the merge base) between base and target commits, and get the diff b</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/diffs/commits?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">importRequestsCreate(organization, body, project, repositoryId, apiVersion, callback)</td>
    <td style="padding:15px">Create an import request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/importRequests?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">importRequestsQuery(organization, project, repositoryId, includeAbandoned, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve import requests for a repository.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/importRequests?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">importRequestsGet(organization, project, repositoryId, importRequestId, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve a particular import request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/importRequests/{pathv4}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">importRequestsUpdate(organization, body, project, repositoryId, importRequestId, apiVersion, callback)</td>
    <td style="padding:15px">Retry or abandon a failed import request.

There can only be one active import request associated w</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/importRequests/{pathv4}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">itemsList(organization, repositoryId, project, itemPath, scopePath, recursionLevel = 'none', includeContentMetadata, latestProcessedChange, download, includeLinks, includeContent, resolveLfs, format, versionDescriptorVersion, versionDescriptorVersionOptions = 'none', versionDescriptorVersionType = 'branch', apiVersion, callback)</td>
    <td style="padding:15px">Get Item Metadata and/or Content for a collection of items. The download parameter is to indicate w</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/items?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">itemsGetItemsBatch(organization, body, repositoryId, project, apiVersion, callback)</td>
    <td style="padding:15px">Post for retrieving a creating a batch out of a set of items in a repo / project given a list of pa</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/itemsbatch?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestQueryGet(organization, body, repositoryId, project, apiVersion, callback)</td>
    <td style="padding:15px">This API is used to find what pull requests are related to a given commit.  It can be used to eithe</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullrequestquery?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestAttachmentsList(organization, repositoryId, pullRequestId, project, apiVersion, callback)</td>
    <td style="padding:15px">Get a list of files attached to a given pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/attachments?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestAttachmentsCreate(organization, body, fileName, repositoryId, pullRequestId, project, apiVersion, callback)</td>
    <td style="padding:15px">Attach a new file to a pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/attachments/{pathv5}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestAttachmentsDelete(organization, fileName, repositoryId, pullRequestId, project, apiVersion, callback)</td>
    <td style="padding:15px">Delete a pull request attachment.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/attachments/{pathv5}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestAttachmentsGet(organization, fileName, repositoryId, pullRequestId, project, apiVersion, callback)</td>
    <td style="padding:15px">Get the file content of a pull request attachment.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/attachments/{pathv5}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestCommitsGetPullRequestCommits(organization, repositoryId, pullRequestId, project, top, continuationToken, apiVersion, callback)</td>
    <td style="padding:15px">Get the commits for the specified pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/commits?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestCommitsGetPullRequestIterationCommits(organization, repositoryId, pullRequestId, iterationId, project, top, skip, apiVersion, callback)</td>
    <td style="padding:15px">Get the commits for the specified iteration of a pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/iterations/{pathv5}/commits?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestIterationsList(organization, repositoryId, pullRequestId, project, includeCommits, apiVersion, callback)</td>
    <td style="padding:15px">Get the list of iterations for the specified pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/iterations?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestIterationsGet(organization, repositoryId, pullRequestId, iterationId, project, apiVersion, callback)</td>
    <td style="padding:15px">Get the specified iteration for a pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/iterations/{pathv5}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestIterationChangesGet(organization, repositoryId, pullRequestId, iterationId, project, top, skip, compareTo, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve the changes made in a pull request between two iterations.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/iterations/{pathv5}/changes?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestIterationStatusesCreate(organization, body, repositoryId, pullRequestId, iterationId, project, apiVersion, callback)</td>
    <td style="padding:15px">Create a pull request status on the iteration. This operation will have the same result as Create s</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/iterations/{pathv5}/statuses?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestIterationStatusesList(organization, repositoryId, pullRequestId, iterationId, project, apiVersion, callback)</td>
    <td style="padding:15px">Get all the statuses associated with a pull request iteration.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/iterations/{pathv5}/statuses?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestIterationStatusesUpdate(organization, body, repositoryId, pullRequestId, iterationId, project, apiVersion, callback)</td>
    <td style="padding:15px">Update pull request iteration statuses collection. The only supported operation type is `remove`.

</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/iterations/{pathv5}/statuses?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestIterationStatusesDelete(organization, repositoryId, pullRequestId, iterationId, statusId, project, apiVersion, callback)</td>
    <td style="padding:15px">Delete pull request iteration status.

You can remove multiple statuses in one call by using Update</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/iterations/{pathv5}/statuses/{pathv6}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestIterationStatusesGet(organization, repositoryId, pullRequestId, iterationId, statusId, project, apiVersion, callback)</td>
    <td style="padding:15px">Get the specific pull request iteration status by ID. The status ID is unique within the pull reque</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/iterations/{pathv5}/statuses/{pathv6}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestLabelsCreate(organization, body, repositoryId, pullRequestId, project, projectId, apiVersion, callback)</td>
    <td style="padding:15px">Create a label for a specified pull request. The only required field is the name of the new label.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/labels?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestLabelsList(organization, repositoryId, pullRequestId, project, projectId, apiVersion, callback)</td>
    <td style="padding:15px">Get all the labels assigned to a pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/labels?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestLabelsDelete(organization, repositoryId, pullRequestId, labelIdOrName, project, projectId, apiVersion, callback)</td>
    <td style="padding:15px">Removes a label from the set of those assigned to the pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/labels/{pathv5}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestLabelsGet(organization, repositoryId, pullRequestId, labelIdOrName, project, projectId, apiVersion, callback)</td>
    <td style="padding:15px">Retrieves a single label that has been assigned to a pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/labels/{pathv5}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestPropertiesList(organization, repositoryId, pullRequestId, project, apiVersion, callback)</td>
    <td style="padding:15px">Get external properties of the pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/properties?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestPropertiesUpdate(organization, body, repositoryId, pullRequestId, project, apiVersion, callback)</td>
    <td style="padding:15px">Create or update pull request external properties. The patch operation can be `add`, `replace` or `</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/properties?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestReviewersCreatePullRequestReviewers(organization, body, repositoryId, pullRequestId, project, apiVersion, callback)</td>
    <td style="padding:15px">Add reviewers to a pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/reviewers?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestReviewersCreateUnmaterializedPullRequestReviewer(organization, body, repositoryId, pullRequestId, project, apiVersion, callback)</td>
    <td style="padding:15px">Add an unmaterialized identity to the reviewers of a pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/reviewers?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestReviewersList(organization, repositoryId, pullRequestId, project, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve the reviewers for a pull request</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/reviewers?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestReviewersUpdatePullRequestReviewers(organization, body, repositoryId, pullRequestId, project, apiVersion, callback)</td>
    <td style="padding:15px">Reset the votes of multiple reviewers on a pull request.  NOTE: This endpoint only supports updatin</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/reviewers?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestReviewersCreatePullRequestReviewer(organization, body, repositoryId, pullRequestId, reviewerId, project, apiVersion, callback)</td>
    <td style="padding:15px">Add a reviewer to a pull request or cast a vote.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/reviewers/{pathv5}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestReviewersDelete(organization, repositoryId, pullRequestId, reviewerId, project, apiVersion, callback)</td>
    <td style="padding:15px">Remove a reviewer from a pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/reviewers/{pathv5}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestReviewersGet(organization, repositoryId, pullRequestId, reviewerId, project, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve information about a particular reviewer on a pull request</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/reviewers/{pathv5}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestReviewersUpdatePullRequestReviewer(organization, body, repositoryId, pullRequestId, reviewerId, project, apiVersion, callback)</td>
    <td style="padding:15px">Edit a reviewer entry. These fields are patchable: isFlagged, hasDeclined</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/reviewers/{pathv5}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestShareSharePullRequest(organization, body, repositoryId, pullRequestId, project, apiVersion, callback)</td>
    <td style="padding:15px">Sends an e-mail notification about a specific pull request to a set of recipients</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/share?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestStatusesCreate(organization, body, repositoryId, pullRequestId, project, apiVersion, callback)</td>
    <td style="padding:15px">Create a pull request status.

The only required field for the status is `Context.Name` that unique</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/statuses?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestStatusesList(organization, repositoryId, pullRequestId, project, apiVersion, callback)</td>
    <td style="padding:15px">Get all the statuses associated with a pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/statuses?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestStatusesUpdate(organization, body, repositoryId, pullRequestId, project, apiVersion, callback)</td>
    <td style="padding:15px">Update pull request statuses collection. The only supported operation type is `remove`.

This opera</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/statuses?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestStatusesDelete(organization, repositoryId, pullRequestId, statusId, project, apiVersion, callback)</td>
    <td style="padding:15px">Delete pull request status.

You can remove multiple statuses in one call by using Update operation.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/statuses/{pathv5}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestStatusesGet(organization, repositoryId, pullRequestId, statusId, project, apiVersion, callback)</td>
    <td style="padding:15px">Get the specific pull request status by ID. The status ID is unique within the pull request across</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/statuses/{pathv5}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestThreadsCreate(organization, body, repositoryId, pullRequestId, project, apiVersion, callback)</td>
    <td style="padding:15px">Create a thread in a pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/threads?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestThreadsList(organization, repositoryId, pullRequestId, project, iteration, baseIteration, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve all threads in a pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/threads?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestThreadsGet(organization, repositoryId, pullRequestId, threadId, project, iteration, baseIteration, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve a thread in a pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/threads/{pathv5}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestThreadsUpdate(organization, body, repositoryId, pullRequestId, threadId, project, apiVersion, callback)</td>
    <td style="padding:15px">Update a thread in a pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/threads/{pathv5}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestThreadCommentsCreate(organization, body, repositoryId, pullRequestId, threadId, project, apiVersion, callback)</td>
    <td style="padding:15px">Create a comment on a specific thread in a pull request (up to 500 comments can be created per thre</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/threads/{pathv5}/comments?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestThreadCommentsList(organization, repositoryId, pullRequestId, threadId, project, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve all comments associated with a specific thread in a pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/threads/{pathv5}/comments?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestThreadCommentsDelete(organization, repositoryId, pullRequestId, threadId, commentId, project, apiVersion, callback)</td>
    <td style="padding:15px">Delete a comment associated with a specific thread in a pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/threads/{pathv5}/comments/{pathv6}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestThreadCommentsGet(organization, repositoryId, pullRequestId, threadId, commentId, project, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve a comment associated with a specific thread in a pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/threads/{pathv5}/comments/{pathv6}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestThreadCommentsUpdate(organization, body, repositoryId, pullRequestId, threadId, commentId, project, apiVersion, callback)</td>
    <td style="padding:15px">Update a comment associated with a specific thread in a pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/threads/{pathv5}/comments/{pathv6}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestCommentLikesCreate(organization, repositoryId, pullRequestId, threadId, commentId, project, apiVersion, callback)</td>
    <td style="padding:15px">Add a like on a comment.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/threads/{pathv5}/comments/{pathv6}/likes?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestCommentLikesDelete(organization, repositoryId, pullRequestId, threadId, commentId, project, apiVersion, callback)</td>
    <td style="padding:15px">Delete a like on a comment.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/threads/{pathv5}/comments/{pathv6}/likes?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestCommentLikesList(organization, repositoryId, pullRequestId, threadId, commentId, project, apiVersion, callback)</td>
    <td style="padding:15px">Get likes for a comment.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/threads/{pathv5}/comments/{pathv6}/likes?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pullRequestWorkItemsList(organization, repositoryId, pullRequestId, project, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve a list of work items associated with a pull request.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pullRequests/{pathv4}/workitems?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pushesCreate(organization, body, repositoryId, project, apiVersion, callback)</td>
    <td style="padding:15px">Push changes to the repository.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pushes?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pushesList(organization, repositoryId, project, skip, top, searchCriteriaFromDate, searchCriteriaIncludeLinks, searchCriteriaIncludeRefUpdates, searchCriteriaPusherId, searchCriteriaRefName, searchCriteriaToDate, apiVersion, callback)</td>
    <td style="padding:15px">Retrieves pushes associated with the specified repository.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pushes?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">pushesGet(organization, repositoryId, pushId, project, includeCommits, includeRefUpdates, apiVersion, callback)</td>
    <td style="padding:15px">Retrieves a particular push.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/pushes/{pathv4}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">refsList(organization, repositoryId, project, filter, includeLinks, includeStatuses, includeMyBranches, latestStatusesOnly, peelTags, filterContains, top, continuationToken, apiVersion, callback)</td>
    <td style="padding:15px">Queries the provided repository for its refs and returns them.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/refs?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">refsUpdateRef(organization, body, repositoryId, filter, project, projectId, apiVersion, callback)</td>
    <td style="padding:15px">Lock or Unlock a branch.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/refs?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">refsUpdateRefs(organization, body, repositoryId, project, projectId, apiVersion, callback)</td>
    <td style="padding:15px">Creating, updating, or deleting refs(branches).

Updating a ref means making it point at a differen</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/refs?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">revertsCreate(organization, body, project, repositoryId, apiVersion, callback)</td>
    <td style="padding:15px">Starts the operation to create a new branch which reverts changes introduced by either a specific c</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/reverts?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">revertsGetRevertForRefName(organization, project, repositoryId, refName, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve information about a revert operation for a specific branch.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/reverts?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">revertsGetRevert(organization, project, revertId, repositoryId, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve information about a revert operation by revert Id.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/reverts/{pathv4}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">statsList(organization, repositoryId, project, baseVersionDescriptorVersion, baseVersionDescriptorVersionOptions = 'none', baseVersionDescriptorVersionType = 'branch', apiVersion, callback)</td>
    <td style="padding:15px">Retrieve statistics about all branches within a repository.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/stats/branches?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">suggestionsList(organization, repositoryId, project, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve a pull request suggestion for a particular repository or team project.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/suggestions?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">treesGet(organization, repositoryId, sha1, project, projectId, recursive, fileName, format, apiVersion, callback)</td>
    <td style="padding:15px">The Tree endpoint returns the collection of objects underneath the specified tree. Trees are folder</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/trees/{pathv4}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">mergeBasesList(organization, repositoryNameOrId, commitId, otherCommitId, project, otherCollectionId, otherRepositoryId, apiVersion, callback)</td>
    <td style="padding:15px">Find the merge bases of two commits, optionally across forks. If otherRepositoryId is not specified</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/commits/{pathv4}/mergebases?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">forksList(organization, repositoryNameOrId, collectionId, project, includeLinks, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve all forks of a repository in the collection.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/forks/{pathv4}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">forksCreateForkSyncRequest(organization, body, repositoryNameOrId, project, includeLinks, apiVersion, callback)</td>
    <td style="padding:15px">Request that another repository's refs be fetched into this one. It syncs two existing forks. To cr</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/forkSyncRequests?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">forksGetForkSyncRequests(organization, repositoryNameOrId, project, includeAbandoned, includeLinks, apiVersion, callback)</td>
    <td style="padding:15px">Retrieve all requested fork sync operations on this repository.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/forkSyncRequests?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">forksGetForkSyncRequest(organization, repositoryNameOrId, forkSyncOperationId, project, includeLinks, apiVersion, callback)</td>
    <td style="padding:15px">Get a specific fork sync operation's details.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/forkSyncRequests/{pathv4}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">mergesCreate(organization, body, project, repositoryNameOrId, includeLinks, apiVersion, callback)</td>
    <td style="padding:15px">Request a git merge operation. Currently we support merging only 2 commits.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/merges?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
  <tr>
    <td style="padding:15px">mergesGet(organization, project, repositoryNameOrId, mergeOperationId, includeLinks, apiVersion, callback)</td>
    <td style="padding:15px">Get a specific merge operation's details.</td>
    <td style="padding:15px">{base_path}/{version}/{pathv1}/{pathv2}/_apis/git/repositories/{pathv3}/merges/{pathv4}?{query}</td>
    <td style="padding:15px">Yes</td>
  </tr>
</table>
<br>

### Authentication

This document will go through the steps for authenticating the Microsoft Azure DevOps adapter with the authentication methods we have worked with in the past. Properly configuring the properties for an adapter in IAP is critical for getting the adapter online. You can read more about adapter authentication <a href="https://docs.itential.com/opensource/docs/authentication" target="_blank">HERE</a>. 

#### Static Token Authentication
The Microsoft Azure DevOps adapter requires Static Token Authentication. If you change authentication methods, you should change this section accordingly and merge it back into the adapter repository.

STEPS  
1. Ensure you have access to a Microsoft Azure DevOps server and that it is running
2. Follow the steps in the README.md to import the adapter into IAP if you have not already done so
3. Use the properties below for the ```properties.authentication``` field
```json
"authentication": {
  "auth_method": "static_token",
  "token": "<PAT>",
  "token_timeout": 600000,
  "token_cache": "local",
  "invalid_token_error": 401,
  "auth_field": "header.headers.Authorization",
  "auth_field_format": "Basic {b64}Basic:{token}{/b64}",
  "auth_logging": false
}
```
you can leave all of the other properties in the authentication section, they will not be used when the auth_method is static_token.

4. Restart the adapter. If your properties were set correctly, the adapter should go online. 

#### Troubleshooting
- Make sure you copied over the correct username and password.
- Turn on debug level logs for the adapter in IAP Admin Essentials.
- Turn on auth_logging for the adapter in IAP Admin Essentials (adapter properties).
- Investigate the logs - in particular:
  - The FULL REQUEST log to make sure the proper headers are being sent with the request.
  - The FULL BODY log to make sure the payload is accurate.
  - The CALL RETURN log to see what the other system is telling us.
- Credentials should be ** masked ** by the adapter so make sure you verify the username and password - including that there are erroneous spaces at the front or end.
- Remember when you are done to turn auth_logging off as you do not want to log credentials.

## Additional Information

### Enhancements

#### Adding a Second Instance of an Adapter

You can add a second instance of this adapter without adding new code on the file system. To do this go into the IAP Admin Essentials and add a new service config for this adapter. The two instances of the adapter should have unique ids. In addition, they should point to different instances (unique host and port) of the other system.

#### Adding Adapter Calls

There are multiple ways to add calls to an existing adapter.

The easiest way would be to use the Adapter Builder update process. This process takes in a Swagger or OpenAPI document, allows you to select the calls you want to add and then generates a zip file that can be used to update the adapter. Once you have the zip file simply put it in the adapter directory and execute `npm run adapter:update`.

```bash
mv updatePackage.zip adapter-azure_devops
cd adapter-azure_devops
npm run adapter:update
```

If you do not have a Swagger or OpenAPI document, you can use a Postman Collection and convert that to an OpenAPI document using APIMatic and then follow the first process.

If you want to manually update the adapter that can also be done the key thing is to make sure you update all of the right files. Within the entities directory you will find 1 or more entities. You can create a new entity or add to an existing entity. Each entity has an action.json file, any new call will need to be put in the action.json file. It will also need to be added to the enum for the ph_request_type in the appropriate schema files. Once this configuration is complete you will need to add the call to the adapter.js file and, in order to make it available as a workflow task in IAP, it should also be added to the pronghorn.json file. You can optionally add it to the unit and integration test files. There is more information on how to work on each of these files in the <a href="https://docs.itential.com/opensource/docs/adapters" target="_blank">Adapter Technical Resources</a> on our Documentation Site.

```text
Files to update
* entities/<entity>/action.json: add an action
* entities/<entity>/schema.json (or the schema defined on the action): add action to the enum for ph_request_type
* adapter.js: add the new method and make sure it calls the proper entity and action
* pronghorn.json: add the new method
* test/unit/adapterTestUnit.js (optional but best practice): add unit test(s) - function is there, any required parameters error when not passed in
* test/integration/adapterTestIntegration.js (optional but best practice): add integration test
```

#### Adding Adapter Properties

While changing adapter properties is done in the service instance configuration section of IAP, adding properties has to be done in the adapter. To add a property you should edit the propertiesSchema.json with the proper information for the property. In addition, you should modify the sampleProperties to have the new property in it.

```text
Files to update
* propertiesSchema.json: add the new property and how it is defined
* sampleProperties: add the new property with a default value
* test/unit/adapterTestUnit.js (optional but best practice): add the property to the global properties
* test/integration/adapterTestIntegration.js (optional but best practice): add the property to the global properties
```

#### Changing Adapter Authentication

Often an adapter is built before knowing the authentication and authentication processes can also change over time. The adapter supports many different kinds of authentication but it does require configuration. Some forms of authentication can be defined entirely with the adapter properties but others require configuration.

```text
Files to update
* entities/.system/action.json: change the getToken action as needed
* entities/.system/schemaTokenReq.json: add input parameters (external name is name in other system)
* entities/.system/schemaTokenResp.json: add response parameters (external name is name in other system)
* propertiesSchema.json: add any new property and how it is defined
* sampleProperties: add any new property with a default value
* test/unit/adapterTestUnit.js (optional but best practice): add the property to the global properties
* test/integration/adapterTestIntegration.js (optional but best practice): add the property to the global properties
```

#### Enhancing Adapter Integration Tests

The adapter integration tests are written to be able to test in either stub (standalone) mode or integrated to the other system. However, if integrating to the other system, you may need to provide better data than what the adapter provides by default as that data is likely to fail for create and update. To provide better data, edit the adapter integration test file. Make sure you do not remove the marker and keep custom code below the marker so you do not impact future migrations. Once the edits are complete, run the integration test as it instructs you to above. When you run integrated to the other system, you can also save mockdata for future use by changing the isSaveMockData flag to true.

```text
Files to update
* test/integration/adapterTestIntegration.js: add better data for the create and update calls so that they will not fail.
```

As mentioned previously, for most of these changes as well as other possible changes, there is more information on how to work on an adapter in the <a href="https://docs.itential.com/opensource/docs/adapters" target="_blank">Adapter Technical Resources</a> on our Documentation Site.

### Contributing

First off, thanks for taking the time to contribute!

The following is a set of rules for contributing.

#### Code of Conduct

This project and everyone participating in it is governed by the Code of Conduct. By participating, you are expected to uphold this code. Please report unacceptable behavior to support@itential.com.

#### How to Contribute

Follow the contributing guide (here)[https://gitlab.com/itentialopensource/adapters/contributing-guide]

### Helpful Links

<a href="https://docs.itential.com/opensource/docs/adapters" target="_blank">Adapter Technical Resources</a>

### Node Scripts

There are several node scripts that now accompany the adapter. These scripts are provided to make several activities easier. Many of these scripts can have issues with different versions of IAP as they have dependencies on IAP and Mongo. If you have issues with the scripts please report them to the Itential Adapter Team. Each of these scripts are described below.

<table border="1" class="bordered-table">
  <tr>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Run</span></th>
    <th bgcolor="lightgrey" style="padding:15px"><span style="font-size:12.0pt">Description</span></th>
  </tr>
  <tr>
    <td style="padding:15px">npm run adapter:install</td>
    <td style="padding:15px">Provides an easier way to install the adapter.</td>
  </tr>
  <tr>
    <td style="padding:15px">npm run adapter:checkMigrate</td>
    <td style="padding:15px">Checks whether your adapter can and should be migrated to the latest foundation.</td>
  </tr>
  <tr>
    <td style="padding:15px">npm run adapter:findPath</td>
    <td style="padding:15px">Can be used to see if the adapter supports a particular API call.</td>
  </tr>
  <tr>
    <td style="padding:15px">npm run adapter:migrate</td>
    <td style="padding:15px">Provides an easier way to update your adapter after you download the migration zip from Itential DevSite.</td>
  </tr>
  <tr>
    <td style="padding:15px">npm run adapter:update</td>
    <td style="padding:15px">Provides an easier way to update your adapter after you download the update zip from Itential DevSite.</td>
  </tr>
  <tr>
    <td style="padding:15px">npm run adapter:revert</td>
    <td style="padding:15px">Allows you to revert after a migration or update if it resulted in issues.</td>
  </tr>
  <tr>
    <td style="padding:15px">npm run troubleshoot</td>
    <td style="padding:15px">Provides a way to troubleshoot the adapter - runs connectivity, healthcheck and basic get.</td>
  </tr>
  <tr>
    <td style="padding:15px">npm run connectivity</td>
    <td style="padding:15px">Provides a connectivity check to the Servicenow system.</td>
  </tr>
  <tr>
    <td style="padding:15px">npm run healthcheck</td>
    <td style="padding:15px">Checks whether the configured healthcheck call works to Servicenow.</td>
  </tr>
  <tr>
    <td style="padding:15px">npm run basicget</td>
    <td style="padding:15px">Checks whether the basic get calls works to Servicenow.</td>
  </tr>
</table>
<br>

## Troubleshoot

Run `npm run troubleshoot` to start the interactive troubleshooting process. The command allows you to verify and update connection, authentication as well as healthcheck configuration. After that it will test these properties by sending HTTP request to the endpoint. If the tests pass, it will persist these changes into IAP.

You also have the option to run individual commands to perform specific test:

- `npm run healthcheck` will perform a healthcheck request of with current setting.
- `npm run basicget` will perform some non-parameter GET request with current setting.
- `npm run connectivity` will perform networking diagnostics of the adatper endpoint.

### Connectivity Issues

1. You can run the adapter troubleshooting script which will check connectivity, run the healthcheck and run basic get calls.

```bash
npm run troubleshoot
```

2. Verify the adapter properties are set up correctly.

```text
Go into the Itential Platform GUI and verify/update the properties
```

3. Verify there is connectivity between the Itential Platform Server and Azure_devops Server.

```text
ping the ip address of Azure_devops server
try telnet to the ip address port of Azure_devops
execute a curl command to the other system
```

4. Verify the credentials provided for Azure_devops.

```text
login to Azure_devops using the provided credentials
```

5. Verify the API of the call utilized for Azure_devops Healthcheck.

```text
Go into the Itential Platform GUI and verify/update the properties
```

### Functional Issues

Adapter logs are located in `/var/log/pronghorn`. In older releases of the Itential Platform, there is a `pronghorn.log` file which contains logs for all of the Itential Platform. In newer versions, adapters can be configured to log into their own files.


